
/* 	This is the base class from which my other applets are derived. 
	That means that this routine declares a lot of functions
	that don't change, and some others that do, and applets that
	are descended from this one need only declare the routines that
	they want to alter. 
	
	It's a heartbreaking amount of code, I know. Look at, say,
	vine2.java, at how much smaller it is. If you can just understand
	how that code relates to this code, you can roll your own
	Java toy. How fine would that be?
	
	This applet is meant to run inside a browser's web page (duh).
	The web page, when it calls or evokes or whatever this applet,
	provides 10 parameters:
			<param name=imageX 			value="400">
			<param name=imageY 			value="499">
				The size of the image to be loaded, in pixels.
			<param name=imageXoffset 	value="50">
			<param name=imageYoffset 	value="50">
				How much of a margin to put at the top and left, in pix.
			<param name=imageXscale 	value="2">
			<param name=imageYscale 	value="2">
				How much to enlarge the image. Positive ints only
			<param name=theImage 		value="water.jpg">
				The image to be loaded by the applet
			<param name=slowdownFactor 	value="1">
				This higher this number is, the slower the applet runs.
				Less than 0 does you no good. 
			<param name=NUM_PARTS 		value="40">
				The maximum number of particles the applet can use.
			
			<param name=backgroundR 	value="0">
			<param name=backgroundG 	value="0">
			<param name=backgroundB 	value="0">
				Background color values, range 0-255. 
				
	This applet loads an image, turns it into a list of integers, 
	and then draws the image using particles. 
	
	It's mainly about 6 functions:
	initAll -- responsible for calling all the initialization routines. 
		It's called only at startup.
	initPart -- sets the variables for a single particle. Called any time
		a particle is created.
	
	drawAll -- draws everything. Called every time the onscreen image is to change.
	drawPArt -- draws a particle. Called a lot more often!
	
	doMove -- called every time we update particles. Mainly a loop for
	movePart -- which moves a single particle. 
		
	
	The first function to be called is start. It makes a thread that calls
	initAll(), and then calls start(). initAll() allocates everything and 
	start() causes run() to be called every so often. run() calls doMove(),
	see above, and then update(), which calls paint(), somehow, 
	which calls drawAll().
	
	If you don't follow that, exactly, it doesn't matter. All you need to know, 
	really, is that I've written two functions, doMove() and drawAll().
	doMove() moves particles and drawAll() draws them. You can learn 
	100000 pages of insider Java poop and you'll find out that all you ever 
	needed to do was move and draw, so skip all that and concentrate
	on what to draw and how to move. 
	
	This is a Java 1.1 applet. 1.1 is not the current Java, but Sun
	makes all versions of Java available a java.sun.com. They 
	sometimes claim that the Javas are backwards-compatible. OK.
	
	Anyway. This is my bit for Java toy-making. I am Neal McDonald,
	this is my code, take and eat, but leave in a pointer to me:
	www.workly.com
	*/

import java.awt.*;
import java.awt.image.*;
import java.net.*;
import java.awt.Font;

public class water extends java.applet.Applet implements Runnable {
	// first, data that child applets might change
	
	// the particle data is all in one big array, parts[][].
	// these variables are constants. I use them to label 
	 static int XPOS=1;			// position
	 static int YPOS=2;
	 static int DIR=3;				// direction of travel
	 static int UPD=4;				// UPDate-- how long till we change colors
	 static int OX=5;				// previous position
	 static int OY=6;
	 static int JUMPED=7;			// whether the particle has skipped over to the 
										// other side of the screen by going out of bounds
	final static int SPEED=8;			// how fast the particle goes
	
	int NUM_PARTS;					// number of particles in the simulation
	static int NUM_FLOATS=20;		// other dimension of parts[][]
									// (also, the max number of variables per particle)
	double parts[][];		// ah-ha
	Color partColors[];		// a place to keep a color for each particle. 


	// next, variables less likely to change; many either set from 
	// html parameters or globals needed to do applet plumbing. 
	
	// variables describing the drawing area
	double MAX_X, MAX_Y, MIN_X, MIN_Y;
	Image offScreenImage;
	Graphics gBuf;
	MediaTracker tracker;
	
	int imageX, imageY; 			// picture X & Y resolution in pixels
	int imageXoffset, imageYoffset; // the left and top margins 
	double imageXscale, imageYscale;	// %age to enlarge the image
	int colorGrid[][][]; 			// one r, g, b for each pixel in the image we load. 
	int thePix[];
	Color backgroundColor;			// canvas color 
	int doErase, doReInit;			// whether to erase image, whether to reinit all parts
	
	// Stuff about thread management and when to run a simulation tick
	Thread gameThread;
	int simTicks, imageStatus; 		// the number of times we've moved the particles
	int slowdownFactor, slowdownCounter;
	int mouseStop;					// allows us to pause the simulation
	
	// math speedups -- looking up a double is a lot fater than Math.sin().
	// OTOH, this might be an old habit dying hard. 
	protected double sines[];
	protected double cosines[];
	
	
	// ////////////////////////////////////////////// Helper functions	

	
	// given a position, return the brightness from that position 
	// on the picture. Return black if there is no color.
	int getColorG(int x, int y) {
		int cx, cy, b;
		b=0; 
		cx = (int)((x-imageXoffset)/imageXscale);
		if ((cx>0) && (cx<imageX)) {
			cy = (int)((y-imageYoffset)/imageYscale);
			if ((cy>0) && (cy<imageY)) {
				b = colorGrid[cx][cy][0];
			}
		}
		return b;
	}
	
	// given a position, return the color from that position 
	// on the picture. Return black if there is no color.
	Color getColorC(int x, int y) {
		int cx, cy;
		Color result = backgroundColor;
		cx = (int)((x - imageXoffset)/imageXscale);
		if ((cx>0) && (cx<imageX)) {
			cy = (int)((y - imageYoffset)/imageYscale);
			if ((cy>0) && (cy<imageY)) {
				result = new Color(colorGrid[cx][cy][0], 
					colorGrid[cx][cy][1], colorGrid[cx][cy][2]);
			}
		}
		return result;
	}


	// use these routines to draw, not because they make anything easier,
	// but because they'll allow us to redirect output. 
	// For instance, to PostScript.
	void r_setColor(Graphics g, Color c) {
		//System.out.println("setColor");
		g.setColor(c);
	}
	void r_fillRect(Graphics g, int x, int y, int w, int h) {
		g.fillRect(x, y, w, h);
	}
	void r_drawLine(Graphics g, int x1, int y1, int x2, int y2) {
		//System.out.println("lineTo");
		g.drawLine(x1, y1, x2, y2);
	}



	// /////////////////////////////////////////////// Initialization functions

	// Set up particle number "which" to be somewhere on the image,
	// with some random direction & lifespan
	void initPart(int which) {
		int c;
		double gx, gy;
		
		gx = (Math.random() * (double)imageX * imageXscale) + (double)imageXoffset;
		gy = (Math.random() * (double)imageY * imageYscale) + (double)imageYoffset;
		
		parts[which][XPOS] = gx;
		parts[which][YPOS] = gy;
		parts[which][OX] = gx;
		parts[which][OY] = gy;
		parts[which][DIR] = Math.random() *360;
		parts[which][SPEED] = 5.0;
		parts[which][UPD] = Math.random() *20;
	
		partColors[which] = getColorC((int)gx, (int)gy);
	}






	/* this is a function that does nothing, but child classes 
		can override it to do any initializations not carried out
		now by initAll-- for instance, this is a way to add
		parameters. */
	void initMore() {
	}


	/* this gets called once, at the beginning. */
	void initAll(int numParts, int numFloats) {
		int i, j, k, r, g, b, n;
		double x;
		String d;
		Image theImage;
		ImageProducer prod;
		ImageObserver dick;
		
		initMore();
		// limits on particle motion
		MAX_X = (double)size().width+20;
		MIN_X = (double)-20;
		MAX_Y = (double)size().height+20;
		MIN_Y = (double)-20;

		simTicks = 0;
		mouseStop = 1;
		// set these variables from parameters in water.html that have the same names
		slowdownFactor = java.lang.Integer.parseInt(getParameter("slowdownFactor"));
		slowdownCounter = 0;

		// set the background color
		r = java.lang.Integer.parseInt(getParameter("backgroundR"));
		g = java.lang.Integer.parseInt(getParameter("backgroundG"));
		b = java.lang.Integer.parseInt(getParameter("backgroundB"));
		backgroundColor = new Color(r, g, b);
		
		// init the sines/cosines arrays
		sines = new double[360];
		cosines = new double[360];
		for (i=0; i<360; ++i) {
			sines[i] = Math.sin((double)(i)/57.295827);
			cosines[i] = Math.cos((double)(i)/57.295827);
		}
	
		// set these variables from parameters in water.html that have the same names
		imageX = java.lang.Integer.parseInt(getParameter("imageX"));
		imageY = java.lang.Integer.parseInt(getParameter("imageY"));
		imageXoffset = java.lang.Integer.parseInt(getParameter("imageXoffset"));
		imageYoffset = java.lang.Integer.parseInt(getParameter("imageYoffset"));
		imageXscale = 0.01 * (double)java.lang.Integer.parseInt(getParameter("imageXscale"));
		imageYscale = 0.01 * (double)java.lang.Integer.parseInt(getParameter("imageYscale"));
		imageStatus = 0;
		// load the image 		
		tracker=new MediaTracker(this);
		theImage = getImage(getCodeBase(),getParameter("theImage"));
		tracker.addImage(theImage,0);
		
		try	{  // wait until the image actually shows up
			tracker.waitForID(0);
		} catch(InterruptedException e) { imageStatus=0; }
		
		//imageY = theImage.getHeight(dick);
		//imageX = theImage.getWidth(dick);
		
		// convert the image to an array of integers, 0xAARRGGBB
		thePix = new int[imageX*imageY];
		PixelGrabber apg = new PixelGrabber(theImage, 0,0, imageX,imageY, thePix, 0, imageX);
		try {
			apg.grabPixels();
		} catch (Exception hell) {
		}
		
		// space for the processed image
		colorGrid = new int[imageX][imageY][3];
		for (i=0; i<imageY; i++) {
			for (j=0; j<imageX; ++j) {
				k = thePix[(i*imageX)+j];
				// just copy the color over. you could do other things...
				colorGrid[j][i][2] = k & 0x000000ff;
				colorGrid[j][i][1] = (k & 0x0000ff00)>>8;
				colorGrid[j][i][0] = (k & 0x00ff0000)>>16;
			}
		}
	

		// ah-ha! init the parts array & the colors	
		parts = new double[numParts][numFloats];
		partColors = new Color[numParts];
		for (i=0; i<numParts; ++i) {
			initPart(i);
		}
		
		mouseStop = 0;
		doErase=1;
		doReInit=0;
	}

	


	// this routine draws particles. 
	// if parts[which][UPD]<0, the part changes color
	// to match whatever the image color is at its location.
	// otherwise, --UPD.
	// hint: there are other ways to draw particles
	void drawPart(Graphics g, int which) {
		int px,py,ox,oy,  cx, cy, b;
		
		px = ((int)parts[which][XPOS]);
		py = ((int)parts[which][YPOS]);				
		ox = ((int)parts[which][OX]);				
		oy = ((int)parts[which][OY]);				
		
		if (parts[which][UPD] <0.0) {
			parts[which][UPD] = 4.0;
			partColors[which] = getColorC(px,py);
		} else { parts[which][UPD] -= 1.0; }
		
		// if you draw right after jumping, you get a line
		// all the way across the screen 
		if (parts[which][JUMPED]<0.01) {
			r_setColor(g, partColors[which]);
			r_drawLine(g, px,py,ox,oy); 
		}
	}
		
	
	void drawAll(Graphics g, int howMany){
		int i;
		
		// for the first 12 ticks, do image processing stuff.
		if ((mouseStop==0) && (slowdownCounter<1)) {
			if ((simTicks<10) || (doErase==1)) {
				r_setColor(g, backgroundColor);
				r_fillRect(g,0,0,size().width, size().height);
				doErase=0;
			}
			if (doReInit==1) {
				doReInit = 0;
				for (i=0; i<howMany; ++i) {
					initPart(i);
				}
			}
			// after that, draw particles. 
			for (i=0; i<howMany; ++i) {
				drawPart(g, i);
			}
					
		} 
	}






	// move the particle in the directio indicated by its DIR.
	// if it moves offscreen, move it to the other side.
	
	// hint: change this function
	void movePart(int which) {
		double dx, dy;
		int ix, jumped;
		
		ix = (int)(parts[which][DIR]) % 360;
		if (ix<0) ix+=360;
		// remember where you just were, so you can draw a line to it. 
		parts[which][OX] = parts[which][XPOS];
		parts[which][OY] = parts[which][YPOS];
		parts[which][XPOS] += parts[which][SPEED] * cosines[ix];
		parts[which][YPOS] += parts[which][SPEED] * sines[ix];
		
		parts[which][DIR] += Math.random() * 20.0-10.0;
		
		jumped=0;
		if (parts[which][XPOS] > MAX_X) {
			parts[which][XPOS] = MIN_X;
			jumped=1;
		}
		if (parts[which][XPOS] < MIN_X) {
			parts[which][XPOS] = MAX_X;
			jumped=1;
		}
		if (parts[which][YPOS] > MAX_Y) {
			parts[which][YPOS] = MIN_Y;
			jumped=1;
		}
		if (parts[which][YPOS] < MIN_Y) {
			parts[which][YPOS] = MAX_Y;
			jumped=1;
		}
		// set this so we don't draw a line across the whole screen
		parts[which][JUMPED] = (double)jumped;
	}
	


	void doMove(int howMany) {
		int i;
		if ((mouseStop==0) && (slowdownCounter<1)) {
			for (i=0; i<howMany; ++i) {
				movePart(i);
			}
			++simTicks;
			slowdownCounter = slowdownFactor; 
		} else {
			--slowdownCounter;
		}
	}









/* ************************************************************* */
/* ************************************************************* */
/* ************************************************************* */
/* ************************************************************* */
/* The routines below here will generally be left alone by child applets. */
/* ************************************************************* */
/* ************************************************************* */
/* ************************************************************* */
/* ************************************************************* */

	// All I want, interaction-wise, is to be able to start and stop
	// the animation. These routines get called once each time the events
	// tehy're named after occur. 
	public boolean keyDown(Event event, int keyCode) {
		requestFocus();
		if (keyCode == (int)' ') { // ooo! and also, use space bar to start over
			doErase=1;
		} else if (keyCode==(int)'r') {
			doReInit=1;
		} else {
			if (mouseStop==0) { mouseStop=1; }
			else { mouseStop=0; }
		}
		return false;
	}
		
	public boolean keyUp(Event event, int keyCode) {	
		requestFocus();
		return false;
	}
	// change this function to change how the applet responds to 
	// mouse clicks. for instance, save x and y and use in movePart()
	public boolean mouseDown(java.awt.Event evt, int x, int y) {
		requestFocus();
		if (mouseStop==0) { mouseStop=1; }
		else { mouseStop=0; }
		return false;
	}	
	public boolean mouseDrag(java.awt.Event evt, int x, int y){
		requestFocus();
		return false;
	}	
	public boolean mouseUp(java.awt.Event evt, int x, int y){
		requestFocus();
		return false;
	}	
	// or save this x and y!
	public boolean mouseMove(java.awt.Event evt, int x, int y) {
		requestFocus();
		return false;
	}	

/* ************************************************************* */
/* ************************************************************* */
	/* the routines below this point are set up so 
		I don't have to change them, ever. */

	final public void paint(Graphics g) {
		drawAll(g, NUM_PARTS);
	}

	final public void update(Graphics g) {
		if (gBuf == null) {		
			offScreenImage = createImage(size().width, size().height);
			gBuf = offScreenImage.getGraphics();
		}	
		paint(gBuf);
		g.drawImage(offScreenImage, 0, 0, this);
	}

	final public void run() {
		Thread.currentThread().setPriority(Thread.NORM_PRIORITY);
		while (true) {
			doMove(NUM_PARTS);
			repaint();
			try {
				Thread.sleep(20);
			} catch (Exception hell) {
			}
		}
	}

	final public void start() {
		NUM_PARTS = java.lang.Integer.parseInt(getParameter("NUM_PARTS"));
		gameThread = new Thread(this);
		initAll(NUM_PARTS, NUM_FLOATS);
		gameThread.start();
	}

	final public void stop() {
		gameThread.stop();
		gameThread = null;
	}


}

